package pep;

import static pep.Main.nDistracters;
import static pep.Main.nTargets;
import static pep.Main.nResponses;

/**This Node subclass defines the operations of Episode nodes.*/
public class Episode extends Node{
 /**response time*/        private double rt;
 /**starting time weight*/ public double timeWeight;
 /**rate of connection building*/
 private static double writeConnect = .01;
 /**rate of decay for connections*/
 private static double decayConnect = .25;
 /**array of distracter connections*/
 public double[][] distWeight;
 /**array of target connections*/
 public double[] targWeight;
 /**array of response connections*/
 public double[] respWeight;
 /**number of distracter sets*/
 private static int nSets = nDistracters.length;

 /**Creates a new Episode.
  * @param subject Subject number used for static variable array.*/
 public Episode(int subject){
  super(subject);
  //code need for no-distracter tasks
  if(nDistracters[0]==0) nSets=0;
  threshold = .02;
  //initialize distracter weight array
  distWeight = new double[nSets][];
  for(int set=0; set<nSets; set++)
   distWeight[set] = new double[nDistracters[set]];
  //initialize target weight array
  targWeight = new double[nTargets];
  for(int item=0; item<nTargets; item++)
   targWeight[item] = 0;
  //initialize response weight array
  respWeight = new double[nResponses];
  //set time weight
  timeWeight = .02;
 }

 /**Updates activation for each cycle and searches responses.*/
 public void activate(){
  //skip if nothing meaningful to modify
  if(input>.0001 || activation>.0001){
   update();
   if(activation>threshold) respSearch();
  }
 }

 /**Adds input to node (distracters).
  * @param distIn Input from distracter Input node.
  * @param set Set of Input nodes.
  * @param item Item in Input node set.*/
 public void memSearch(double distIn,int set,int item){
  //compute input
  input += (distIn * distWeight[set][item]);
 }

 /**Adds input to node (targets).
  * @param targIn Input from target Input node.
  * @param item Item in Input node set.*/
 public void memSearch(double targIn,int item){
  //compute input
  input += (targIn * targWeight[item]);
 }

 /**Activates Response nodes.*/
 private void respSearch(){
  //calculate outgoing activation strength
  double temp = (activation-threshold);
  //loop through each Response node
  for(int j=0; j<nResponses; j++){
   //multiply outgoing activation by connection strength
   double output = temp*respWeight[j];
   Main.runs[subj].responses[j].retrieved += output;
   //increment total retrieval
   Response.totRetrieved[subj] += output;
  }
 }

 /**Connects Episode to activated distracter Input nodes.
  * @param distIn Input from distracter Input node.
  * @param set Set of Input nodes.
  * @param item Item in Input node set.*/
 public void connectIn(double distIn,int set,int item){
  //add input to episode
  input += distIn * distWeight[set][item];
  //build connection
  distWeight[set][item] *= (1-writeConnect);
  distWeight[set][item] += (distIn*writeConnect);
 }

 /**Connects Episode to activated target Input nodes.
  * @param targIn Input from target Input node.
  * @param item Item in Input node set.*/
 public void connectIn(double targIn,int item){
  //add input to episode
  input += targIn * targWeight[item];
  //build connection
  targWeight[item] *= (1-writeConnect);
  targWeight[item] += (targIn*writeConnect);
 }

 /**Connects Episode to activated Response nodes.*/
 public void connectOut(){
  //increase activation of new node
  input += .05;
  update();
  //loop through Response nodes
  for(int j=0; j<nResponses; j++){
   //connect to Response node to the extent that it is active
   respWeight[j] *= (1-writeConnect);
   respWeight[j] += Main.runs[subj].responses[j].activation*writeConnect;
  }
 }

 /**Tags an Episode as an error.*/
 public void tagError(){
  //weaken search strength
  double adjust = .02;
  for(int set=0; set<nSets; set++)
   for(int item=0; item<nDistracters[set]; item++)
    distWeight[set][item] *= adjust;
  for(int item=0; item<nTargets; item++)
   targWeight[item] *= adjust;
  for(int item=0; item<nResponses; item++)
   targWeight[item] *= adjust;
  timeWeight *= adjust;
 }

 /**Records response time.
  * @param respTime Response time.*/
 public void setRT(double respTime){
  rt = respTime;
 }

 /**Retrieves response time.
  * @param pace Current pace.
  * @return The amount to change the pace.*/
 public double getRT(double pace){
  double out = 0;
  if(activation>threshold){
   double diff = rt-pace;
   out = diff*timeWeight*(activation-threshold);
  }
  return out;
 }

 /**Adjusts feedback from Response node.*/
 public void feedback(){
  double adjust = range(1-(activation*decayConnect));
  for(int set=0; set<nSets; set++)
   for(int item=0; item<nDistracters[set]; item++)
    distWeight[set][item] *= adjust;
  for(int item=0; item<nTargets; item++)
   targWeight[item] *= adjust;
  for(int item=0; item<nResponses; item++)
   targWeight[item] *= adjust;
  timeWeight *= adjust;
 }
}