package pep;

import java.io.*;                  import java.util.*;
import static pep.Main.nTargets;   import static pep.Main.nDistracters;
import static pep.Main.nResponses; import static pep.Main.isConflict;
import static pep.Main.tSignal;    import static pep.Main.dSignal;
import static pep.Main.tBias;      import static pep.Main.dBias;
import static pep.Main.tOnset;     import static pep.Main.tOffset;
import static pep.Main.dOnset;     import static pep.Main.dOffset;
import static pep.Main.dMask;      import static pep.Main.nConflicting;
import static pep.Main.isi;        import static pep.Main.trialLength;
import static pep.Main.isDegrade;  import static pep.Main.tDegrade;

/**This class is guts of the model, where each run is performed.*/
public class RunTime implements Runnable {
 /**number of distracter sets*/ static int nSets = nDistracters.length;
 /**number of variables*/       static int nVariables = Main.numPerVar.length;
 /**attention node*/            Attention attention;
 /**number of episodes, should be greater than number of trials*/
 static int nEpisodes = 1002;
 /**episodic memory*/
 Episode[] memory = new Episode[nEpisodes];
 /**target Input nodes*/
 Input[] targets = new Input[nTargets];
 /**distracter Input nodes*/
 Input[][] distracters = new Input[nSets][];
 /**Identity nodes*/
 Identity[] identities = new Identity[nTargets];
 /**Response nodes*/
 Response[] responses = new Response[nResponses];

 /**episode counter*/     int eNumber;
 /**end of trial marker*/ boolean isTrialOver;
 /**subject number*/      int subj;
 /**condition number*/    int[] condition = new int[nVariables];
 /**target number*/       int target;
 /**distracter number*/   int[] distracter = new int[nSets];
 /**cycle number*/        int cycle;
 /**correct response*/    int correctResp;
 /**actual response*/     int response;
 /**accuracy*/            int accuracy;
 /**stimulus on screen?*/ boolean isStimuli = true;
 /**end of trials*/       private boolean stop;
 /**number generator*/    private Random r = new Random();
 /**adjustable signal strength*/
  private double[] dSigTrend = (double[])dSignal.clone();
 /**threshold at response*/ private double threshTest = 0;
 /**response feedback*/     private double[] respFB = new double[nResponses];
 double pace;

 /**Sets up the thread.
  * @param subj The subject number.*/
 @SuppressWarnings("empty-statement")
 public RunTime(int subject){
  subj=subject;
 }

 /**Defines the nodes of the model defined and runs the trials.*/
 @Override
 public void run(){
     pace = 2000;
  //code need for no-distracter tasks
  if(nDistracters[0]==0) nSets=0;
  //sets up the nodes for the distracter sets
  for(int set=0; set<nSets; set++){
   distracters[set] = new Input[nDistracters[set]];
   for(int item=0; item<nDistracters[set]; item++){
    distracters[set][item] = new Input(subj);
    distracters[set][item].ignored(nSets);
   }
  }
  //sets up the rest of the nodes
  for(int i=0; i<nEpisodes; i++) memory[i] = new Episode(subj);
  for(int i=0; i<nTargets; i++){
   targets[i] = new Input(subj);
   targets[i].targeted();
   identities[i] = new Identity(subj);
  }
  for(int i=0; i<nResponses; i++) responses[i] = new Response(subj);
  //set episode and connection counters to 0
  eNumber = 0;
  attention = new Attention(subj);

  //---Open I/O Streams---
  StringTokenizer tokenizer = null;
  FileOutputStream out = null;
  PrintStream p = null;
  DataSet trials = new DataSet(Main.source,subj+1);
  try{
   String outFile = "out"+(subj+1)+".txt";
   out = new FileOutputStream(outFile);
   p = new PrintStream(out);
   //get Trial 1
   tokenizer = new StringTokenizer(trials.get(0));
  }
  catch(Exception e){
   System.out.println("Error setting up input and output streams.");
  }

  //---Trial Loop---
  stop = false;
  while(!stop){
   //split up trial info into variables
   correctResp = Integer.parseInt(tokenizer.nextToken());
   target = Integer.parseInt(tokenizer.nextToken());
   for(int ns=0; ns<nSets; ns++)
    distracter[ns] = Integer.parseInt(tokenizer.nextToken());
   for(int nc=0; nc<nVariables; nc++)
    condition[nc] = Integer.parseInt(tokenizer.nextToken());
   response = 999; accuracy = 999;

   //---Attentional Wandering---
   //calculate random change
   double rand = (r.nextDouble() - r.nextDouble());
   //for each distracter set, compute new distracter signal
   for(int j=0; j<nSets; j++)
    dSigTrend[j] = attention.wander(dSigTrend[j],dSignal[j],rand);

   //---Define Inputs and Run Cycles---
   //start with null input
   for(int i=0; i<nTargets; i++) targets[i].shutOff();
   for(int j=0; j<nSets; j++)
    for(int i=0; i<nDistracters[j]; i++) distracters[j][i].shutOff();
   //prepare loop
   isTrialOver = false; isStimuli = true;
   //the presented nodes are activated with noise plus a signal
   for(cycle=1; isTrialOver==false; cycle++){
    //target onset and offset
    if(tOnset==cycle){
     for(int i=0; i<nTargets; i++) targets[i].turnOn(tBias,0);
     targets[target].turnOn(tBias,tSignal);
     // <-- insert asterisk between slashes to lesion degrade check
     if(isDegrade && condition[nVariables-1]==1)
      targets[target].turnOn(tBias,tDegrade);//*/
    }
    else if(tOffset<cycle) targets[target].remove();
    //distracter onset, offset, and masking
    if(dOnset==cycle){
     for(int j=0; j<nSets; j++){
      for(int i=0; i<nDistracters[j]; i++)
       distracters[j][i].turnOn(dBias[j],0);
      distracters[j][distracter[j]].turnOn(dBias[j],dSigTrend[j]);
     }
    }
    else if(dMask==cycle){
     for(int j=0; j<nSets; j++){
      for(int i=0; i<nDistracters[j]; i++)
       distracters[j][i].mask();
     }
    }
    else if(dOffset<cycle){
     for(int j=0; j<nSets; j++){
      for(int i=0; i<nDistracters[j]; i++)
       distracters[j][i].remove();
     }
    }

    //run a cycle and monitor for colour and response identification
    go();
    if(monitorTarg()) if(monitorResp()) isTrialOver=true;
    if((cycle-tOnset)>trialLength-1) isTrialOver=true;
   }

   //---Post-Trial Coding---
   //calculate accuracy
   //correct cycle increment
   cycle--;
   if(correctResp==response){
    accuracy = 1;
    pace = pace*.5 + (cycle-tOnset)*.5;
   }
   else{
    accuracy = 0;
    memory[eNumber].tagError();
   }

   //send feedback signal
   for(int ep=0; ep<eNumber; ep++)
    memory[ep].feedback();
   if(accuracy==1) identities[response].reinforce();

   //write trial info to file
   p.print(""+(subj+1)+"\t"+response+"\t"+target+"\t");
   for(int k=0; k<nSets; k++) p.print(distracter[k]+"\t");
   for(int m=0; m<nVariables; m++) p.print(condition[m]+"\t");
   p.println(correctResp+"\t"+(cycle-tOnset)+"\t"+accuracy+"\t"+threshTest);
   //record response time into memory
   memory[eNumber].setRT(cycle-tOnset);

   //---Run ISI---
   isStimuli = false;
   //null input cycles (inter-trial interval)
   for(int i=0; i<nTargets; i++) targets[i].shutOff();
   for(int j=0; j<nSets; j++)
    for(int i=0; i<nDistracters[j]; i++) distracters[j][i].shutOff();
   //loop through blank cycles
   for(int i=0; i<isi; i++){
    go();
   }

   //increase episode and connection counters
   eNumber++;
   if(eNumber>=nEpisodes) System.out.println("Error: Not enough Episodes.");

   //---Read Next Trial---
   try{
    String tempTrial = trials.get(eNumber);
    if(tempTrial.equals("END_OF_FILE")) stop=true;
    else tokenizer = new StringTokenizer(tempTrial);
   }
   catch(Exception e2){
    System.out.println("No END_OF_FILE marker");
    stop=true;
   }
  }

  //---Exit Simulation---
  //tag end of file
  p.println("END_OF_FILE");
  //close input and output streams
  try{
   p.close(); out.close();
  }
  catch(Exception e3){
   System.out.println("Error closing input and output streams.");
  }

  //release thread to freeThread semaphore to allow a new thread
  Main.freeThread.release();
 }


 //-------------------------------------------------------------------
 //------------------------------Methods------------------------------
 //-------------------------------------------------------------------

 /**Performs all the calculations for an individual cycle.*/
 private void go(){
  //---INPUTS---
  for(int set=0; set<nSets; set++){
   boolean isConf = isConflict;
   //only the first node set conflicts
   if(set>0) isConf=false;
   for(int item=0; item<nDistracters[set]; item++){
    if(item>=nConflicting) isConf=false;
    distracters[set][item].activate(item,isConf);
   }
  }
  for(int i=0; i<nTargets; i++) targets[i].activate(i,true);
  //---EPISODES---
  //loop for distracter connections
  for(int set=0; set<nSets; set++){
   for(int item=0; item<nDistracters[set]; item++){
    double temp = distracters[set][item].getOut();
    if(temp>0){
     //first try larger loop
     int eNumMinus10 = eNumber - 10;
     int eLoop = 0;
     while(eLoop<eNumMinus10){
      memory[eLoop].memSearch(temp,set,item); eLoop++;
      memory[eLoop].memSearch(temp,set,item); eLoop++;
      memory[eLoop].memSearch(temp,set,item); eLoop++;
      memory[eLoop].memSearch(temp,set,item); eLoop++;
      memory[eLoop].memSearch(temp,set,item); eLoop++;
      memory[eLoop].memSearch(temp,set,item); eLoop++;
      memory[eLoop].memSearch(temp,set,item); eLoop++;
      memory[eLoop].memSearch(temp,set,item); eLoop++;
      memory[eLoop].memSearch(temp,set,item); eLoop++;
      memory[eLoop].memSearch(temp,set,item); eLoop++;
     }
     //run remaining nodes
     for(; eLoop<eNumber; eLoop++) memory[eLoop].memSearch(temp,set,item);
    }
    memory[eNumber].connectIn(temp,set,item);
   }
  }
  //loop for target connections
  for(int item=0; item<nTargets; item++){
   double temp = targets[item].getOut();
   if(temp>0){
    //first try larger loop
    int eNumMinus10 = eNumber - 10;
    int eLoop = 0;
    while(eLoop<eNumMinus10){
     memory[eLoop].memSearch(temp,item); eLoop++;
     memory[eLoop].memSearch(temp,item); eLoop++;
     memory[eLoop].memSearch(temp,item); eLoop++;
     memory[eLoop].memSearch(temp,item); eLoop++;
     memory[eLoop].memSearch(temp,item); eLoop++;
     memory[eLoop].memSearch(temp,item); eLoop++;
     memory[eLoop].memSearch(temp,item); eLoop++;
     memory[eLoop].memSearch(temp,item); eLoop++;
     memory[eLoop].memSearch(temp,item); eLoop++;
     memory[eLoop].memSearch(temp,item); eLoop++;
    }
    //run remaining nodes
    for(; eLoop<eNumber; eLoop++) memory[eLoop].memSearch(temp,item);
   }
   memory[eNumber].connectIn(temp,item);
  }
  //first try larger loop
  int eNumMinus10 = eNumber - 10;
  int eLoop = 0;
  while(eLoop<eNumMinus10){
   memory[eLoop].activate(); eLoop++; memory[eLoop].activate(); eLoop++;
   memory[eLoop].activate(); eLoop++; memory[eLoop].activate(); eLoop++;
   memory[eLoop].activate(); eLoop++; memory[eLoop].activate(); eLoop++;
   memory[eLoop].activate(); eLoop++; memory[eLoop].activate(); eLoop++;
   memory[eLoop].activate(); eLoop++; memory[eLoop].activate(); eLoop++;
  }
  //run remaining nodes
  for(; eLoop<eNumber; eLoop++) memory[eLoop].activate();
  //update current episode
  if(isStimuli) memory[eNumber].connectOut();
  //---IDENTITIES---
  for(int i=0; i<nTargets; i++) identities[i].sumAct(subj);
  for(int i=0; i<nTargets; i++) identities[i].activate(i);
  Identity.clear(subj);
  //---RESPONSES---
  for(int i=0; i<nResponses; i++) responses[i].activate();
  Response.clear(subj);
 }

 /**Checks colour identification.
  *@return If target is present.*/
 private boolean monitorTarg(){
  //used to prevent response firing when no stimuli are identified
  boolean found = false;
  for(int i=0; i<nTargets; i++)
   if(targets[i].activation>targets[i].threshold) found = true;
  return found;
 }

 /**Checks for temporal expectancy and generates a temporal bias.
  *@return Temporal expectancy bias.*/
 private double monitorTemp(){
  double expect = 0;
  // <--insert asterisk between slashes to lesion temporal learning 
  //compute amount to adjust pace
  double adjustment = 0;
  for(int ep=0; ep<eNumber; ep++) adjustment += memory[ep].getRT(pace);
  //set current cycle pace
  pace += adjustment;
  //compute threshold
  double match = (cycle-tOnset)-pace;
  expect = .25;
  expect += Math.pow(match,2)*.00002;
  if(expect>.45) expect=.45;
  //*/expect=.35;
  return expect;
 }

 /**Checks for response identification and ends the trial if necessary.
  *@return If the response has been found.*/
 private boolean monitorResp(){
  boolean found = false;
  double currentThresh = monitorTemp();
  for(int i=0; i<nResponses; i++){
   if(responses[i].activation>currentThresh){
    found = true;
    threshTest = currentThresh;
    //if two responses cross threshold at same time, response = 777
    if(response != 999) response = 777;
    else response = i;
   }
  }
  return found;
 }
}